import copy

import numpy as np
import pandas as pd
import pytest

from darts.utils.data.tabularization import add_static_covariates_to_lagged_data
from darts.utils.timeseries_generation import linear_timeseries


class TestAddStaticToLaggedData:
    series = linear_timeseries(length=6)
    series = series.stack(series)
    series_stcov_single = series.with_static_covariates(pd.DataFrame({"a": [0.0]}))
    series_stcov_multi = series.with_static_covariates(
        pd.DataFrame({"a": [0.0], "b": [1.0]})
    )
    series_stcov_multivar = series.with_static_covariates(
        pd.DataFrame({"a": [0.0, 1.0], "b": [10.0, 20.0]})
    )
    features = np.empty(shape=(len(series), 2))

    def test_add_static_covs_train(self):
        # training when `last_shape=None`

        # does not use static covariates -> input = output
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series,
            uses_static_covariates=False,
            last_shape=None,
        )
        assert features.shape == self.features.shape
        assert last_shape is None

        with pytest.raises(ValueError):
            # use static covariates enforces that series have matching static covs
            add_static_covariates_to_lagged_data(
                copy.deepcopy(self.features),
                self.series,
                uses_static_covariates=True,
                last_shape=None,
            )

        with pytest.raises(ValueError):
            # some series with static covs, other without should raise an error
            add_static_covariates_to_lagged_data(
                [copy.deepcopy(self.features), copy.deepcopy(self.features)],
                [self.series, self.series_stcov_single],
                uses_static_covariates=True,
                last_shape=None,
            )

        with pytest.raises(ValueError):
            # mismatching static covs should raise an error
            add_static_covariates_to_lagged_data(
                [copy.deepcopy(self.features), copy.deepcopy(self.features)],
                [self.series_stcov_single, self.series_stcov_multi],
                uses_static_covariates=True,
                last_shape=None,
            )

        # single static cov should yield one added column
        expected_shape = (self.features.shape[0], self.features.shape[1] + 1)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_single,
            uses_static_covariates=True,
            last_shape=None,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_single.static_covariates.shape

        # multi static cov should yield two added columns
        expected_shape = (self.features.shape[0], self.features.shape[1] + 2)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_multi,
            uses_static_covariates=True,
            last_shape=None,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_multi.static_covariates.shape

        # multivariate static cov should yield four added columns
        expected_shape = (self.features.shape[0], self.features.shape[1] + 4)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_multivar,
            uses_static_covariates=True,
            last_shape=None,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_multivar.static_covariates.shape

        # multi series with multivariate static cov should yield four added columns per series
        expected_shape = (self.features.shape[0], self.features.shape[1] + 4)
        features, last_shape = add_static_covariates_to_lagged_data(
            [copy.deepcopy(self.features), copy.deepcopy(self.features)],
            [self.series_stcov_multivar, self.series_stcov_multivar],
            uses_static_covariates=True,
            last_shape=None,
        )
        assert [features_.shape == expected_shape for features_ in features]
        assert last_shape == self.series_stcov_multivar.static_covariates.shape
        assert np.all(
            features[0][:, -sum(last_shape) :] == np.array([0.0, 1.0, 10.0, 20.0])
        )

    def test_add_static_covs_predict(self):
        # predicting when `last_shape` other than `None`

        # does not use static covariates -> input = output
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series,
            uses_static_covariates=False,
            last_shape=(10, 10),
        )
        assert features.shape == self.features.shape
        assert last_shape == (10, 10)

        with pytest.raises(ValueError):
            # last static cov shape is given, but no covariates are available
            add_static_covariates_to_lagged_data(
                copy.deepcopy(self.features),
                self.series,
                uses_static_covariates=True,
                last_shape=(10, 10),
            )

        with pytest.raises(ValueError):
            # when last static cov shape is other than current one, raise an error
            add_static_covariates_to_lagged_data(
                copy.deepcopy(self.features),
                self.series_stcov_single,
                uses_static_covariates=True,
                last_shape=(10, 10),
            )

        # single static cov should yield one added column
        expected_shape = (self.features.shape[0], self.features.shape[1] + 1)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_single,
            uses_static_covariates=True,
            last_shape=self.series_stcov_single.static_covariates.shape,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_single.static_covariates.shape

        # multi static cov should yield two added columns
        expected_shape = (self.features.shape[0], self.features.shape[1] + 2)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_multi,
            uses_static_covariates=True,
            last_shape=self.series_stcov_multi.static_covariates.shape,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_multi.static_covariates.shape

        # multivariate static cov should yield four added columns
        expected_shape = (self.features.shape[0], self.features.shape[1] + 4)
        features, last_shape = add_static_covariates_to_lagged_data(
            copy.deepcopy(self.features),
            self.series_stcov_multivar,
            uses_static_covariates=True,
            last_shape=self.series_stcov_multivar.static_covariates.shape,
        )
        assert features.shape == expected_shape
        assert last_shape == self.series_stcov_multivar.static_covariates.shape

        # multi series with multivariate static cov should yield four added columns per series
        expected_shape = (self.features.shape[0], self.features.shape[1] + 4)
        features, last_shape = add_static_covariates_to_lagged_data(
            [copy.deepcopy(self.features), copy.deepcopy(self.features)],
            [self.series_stcov_multivar, self.series_stcov_multivar],
            uses_static_covariates=True,
            last_shape=self.series_stcov_multivar.static_covariates.shape,
        )
        assert [features_.shape == expected_shape for features_ in features]
        assert last_shape == self.series_stcov_multivar.static_covariates.shape
        assert np.all(
            features[0][:, -sum(last_shape) :] == np.array([0.0, 1.0, 10.0, 20.0])
        )



from collections.abc import Sequence
from enum import Enum
from typing import Optional, Union

import pandas as pd

from darts import TimeSeries


class LikelihoodType(Enum):
    Gaussian = "gaussian"
    Poisson = "poisson"
    NegativeBinomial = "negativebinomial"
    Bernoulli = "bernoulli"
    Beta = "beta"
    Cauchy = "cauchy"
    ContinuousBernoulli = "continuousbernoulli"
    Dirichlet = "dirichlet"
    Exponential = "exponential"
    Gamma = "gamma"
    Geometric = "geometric"
    Gumbel = "gumbel"
    HalfNormal = "halfnormal"
    Laplace = "laplace"
    LogNormal = "lognormal"
    Weibull = "weibull"
    Quantile = "quantile"


class Likelihood:
    def __init__(
        self,
        likelihood_type: LikelihoodType,
        parameter_names: list[str],
    ):
        """
        Base class for all likelihoods.

        * likelihoods for torch models
        * likelihoods for sklearn-like models (e.g. RegressionModel subclasses)

        Parameters
        ----------
        likelihood_type
            A pre-defined `LikelihoodType`.
        parameter_names
            The likelihood (distribution) parameter names.
        """
        self._likelihood_type = likelihood_type
        self._parameter_names = parameter_names

        # used for equality operator between likelihood objects
        self.ignore_attrs_equality = [
            "_likelihood_type",
            "_parameter_names",
            "ignore_attrs_equality",
        ]

    def component_names(self, input_series: TimeSeries) -> list[str]:
        """Generates names for the parameters of the Likelihood."""
        return likelihood_component_names(
            components=input_series.components, parameter_names=self.parameter_names
        )

    @property
    def type(self) -> LikelihoodType:
        """Returns the likelihood type."""
        return self._likelihood_type

    @property
    def parameter_names(self) -> list[str]:
        """Returns the likelihood parameter names."""
        return self._parameter_names

    @property
    def num_parameters(self) -> int:
        """Returns the number of distribution parameters for a single target value."""
        return len(self.parameter_names)

    @staticmethod
    def _get_equality_attrs(likelihood, ignore_attrs):
        # ignore the attributes listed in `ignore_attrs_equality`
        return {k: v for k, v in likelihood.__dict__.items() if k not in ignore_attrs}

    def __eq__(self, other: "Likelihood") -> bool:
        """Defines (in)equality between two likelihood objects."""
        # ignore the attributes listed in `ignore_attrs_equality`
        if type(other) is type(self):
            other_state = self._get_equality_attrs(other, self.ignore_attrs_equality)
            self_state = self._get_equality_attrs(self, self.ignore_attrs_equality)
            return other_state == self_state
        else:
            return False

    def __repr__(self):
        attrs = ", ".join(
            f"{k}={v}"
            for k, v in self._get_equality_attrs(
                self, self.ignore_attrs_equality
            ).items()
        )
        return f"{self.__class__.__name__}({attrs})"


def likelihood_component_names(
    components: Union[pd.Index, list[str]], parameter_names: list[str]
):
    """Generates formatted likelihood parameter names for components and parameter names.

    The order of the returned names is: `[comp1_param_1, ... comp1_param_n, ..., comp_n_param_n]`.

    Parameters
    ----------
    components
        A sequence of component names to add to the beginning of the returned names.
    parameter_names
        A sequence of likelihood parameter names to add to the end of the returned names.
    """
    return [
        f"{tgt_name}_{param_n}"
        for tgt_name in components
        for param_n in parameter_names
    ]


def quantile_names(q: Union[float, list[float]], component: Optional[str] = None):
    """Generates formatted quantile names, optionally added to a component name.

    Parameters
    ----------
    q
        A float or list of floats with the quantiles to generate the names for.
    component
        Optionally, a component name to add to the beginning of the quantile names.
    """
    # predicted quantile text format
    comp = f"{component}_" if component is not None else ""
    if isinstance(q, float):
        return f"{comp}q{q:.2f}"
    else:
        return [f"{comp}q{q_i:.2f}" for q_i in q]


def quantile_interval_names(
    q_interval: Union[tuple[float, float], Sequence[tuple[float, float]]],
    component: Optional[str] = None,
):
    """Generates formatted quantile interval names, optionally added to a component name.

    Parameters
    ----------
    q_interval
        A tuple or multiple tuples with the (lower bound, upper bound) of the quantile intervals.
    component
        Optionally, a component name to add to the beginning of the quantile names.
    """
    # predicted quantile text format
    comp = f"{component}_" if component is not None else ""
    if isinstance(q_interval, tuple):
        return f"{comp}q{q_interval[0]:.2f}_q{q_interval[1]:.2f}"
    else:
        return [f"{comp}q{q_lo:.2f}_q{q_hi:.2f}" for q_lo, q_hi in q_interval]



"""
Additional util functions
-------------------------
"""

from collections.abc import Sequence
from enum import Enum
from functools import total_ordering
from typing import Optional, Union

from darts import TimeSeries
from darts.logging import get_logger, raise_log

try:
    from IPython import get_ipython
except ModuleNotFoundError:
    get_ipython = None

logger = get_logger(__name__)

_SEQ_TYPE_NAMES = {
    0: "`TimeSeries`",
    1: "`Sequence[TimeSeries]`",
    2: "`Sequence[Sequence[TimeSeries]]`",
}


@total_ordering
class SeriesType(Enum):
    """An Enum for different `TimeSeries` sequence types."""

    NONE = -1  # `None`
    SINGLE = 0  # `TimeSeries`
    SEQ = 1  # `Sequence[TimeSeries]`
    SEQ_SEQ = 2  # `Sequence[Sequence[TimeSeries]]`

    def _check_member(self, other):
        if self.__class__ is not other.__class__:
            raise_log(ValueError("`other` must be a `SeriesType` enum."), logger=logger)

    def __eq__(self, other):
        self._check_member(other)
        return super().__eq__(other)

    def __lt__(self, other):
        self._check_member(other)
        return self.value < other.value

    def __add__(self, other: int):
        if not isinstance(other, int):
            raise_log(ValueError("`other` must be of type `int`."), logger=logger)
        new_val = self.value + other
        if new_val > 2:
            raise_log(
                ValueError("Cannot go higher than `SeriesType.SEQ_SEQ`."), logger=logger
            )
        return SeriesType(new_val)

    def __str__(self):
        return _SEQ_TYPE_NAMES[self.value]


def series2seq(
    ts: Optional[
        Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
    ],
    seq_type_out: SeriesType = SeriesType.SEQ,
    nested: bool = False,
) -> Optional[Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]]:
    """If possible, converts `ts` into the desired sequence type `seq_type_out`. Otherwise, returns the
    original `ts`.

    Parameters
    ----------
    ts
        None, a single TimeSeries, a sequence of TimeSeries, or a sequence of sequences of TimeSeries.
    seq_type_out
        The output sequence type:

        - SeriesType.SINGLE: `TimeSeries` (e.g. a single series)
        - SeriesType.SEQ: sequence of `TimeSeries` (e.g. multiple series)
        - SeriesType.SEQ_SEQ: sequence of sequences of `TimeSeries` (e.g. historical forecasts output)
    nested
        Only applies with `seq_type_out=SeriesType.SEQ_SEQ` and `ts` having a sequence type `SeriesType.SEQ`.
        In this case, wrap each element in `ts` in a list ([ts1, ts2] -> [[ts1], [ts2]]).

    Raises
    ------
    ValueError
        If there is an invalid `seq_type_out` value.
    """
    if ts is None:
        return ts

    if not isinstance(seq_type_out, SeriesType):
        raise_log(
            ValueError(
                f"Invalid parameter `seq_type_out={seq_type_out}`. Must be one of `(0, 1, 2)`"
            ),
            logger=logger,
        )

    seq_type_in = get_series_seq_type(ts)

    if seq_type_out == seq_type_in:
        return ts

    n_series = 1 if seq_type_in == SeriesType.SINGLE else len(ts)

    if seq_type_in == SeriesType.SINGLE and seq_type_out == SeriesType.SEQ:
        # ts -> [ts]
        return [ts]
    elif seq_type_in == SeriesType.SINGLE and seq_type_out == SeriesType.SEQ_SEQ:
        # ts -> [[ts]]
        return [[ts]]
    elif (
        seq_type_in == SeriesType.SEQ
        and seq_type_out == SeriesType.SINGLE
        and n_series == 1
    ):
        # [ts] -> ts
        return ts[0]
    elif seq_type_in == SeriesType.SEQ and seq_type_out == SeriesType.SEQ_SEQ:
        if not nested:
            # [ts1, ts2] -> [[ts1, ts2]]
            return [ts]
        else:
            # [ts1, ts2] -> [[ts1], [ts2]]
            return [[ts_] for ts_ in ts]
    elif (
        seq_type_in == SeriesType.SEQ_SEQ
        and seq_type_out == SeriesType.SINGLE
        and n_series == 1
    ):
        # [[ts]] -> [ts]
        return ts[0]
    elif (
        seq_type_in == SeriesType.SEQ_SEQ
        and seq_type_out == SeriesType.SEQ
        and n_series == 1
    ):
        # [[ts1, ts2]] -> [[ts1, ts2]]
        return ts
    else:
        # ts -> ts
        return ts


def seq2series(
    ts: Optional[Union[TimeSeries, Sequence[TimeSeries]]],
) -> Optional[TimeSeries]:
    """If `ts` is a Sequence with only a single series, return the single series as TimeSeries.

    Parameters
    ----------
    ts
        None, a single TimeSeries, or a sequence of TimeSeries

    Returns
    -------
        `ts` if `ts` if is not a single element TimeSeries sequence, else `ts[0]`

    """
    return series2seq(ts, seq_type_out=SeriesType.SINGLE)


def get_single_series(
    ts: Optional[
        Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
    ],
) -> Optional[TimeSeries]:
    """Returns a single (first) TimeSeries or `None` from `ts`. Returns `ts` if  `ts` is a TimeSeries, `ts[0]` if
    `ts` is a `Sequence[TimeSeries]`, and `ts[0][0]` if `ts` is a `Sequence[Sequence[TimeSeries]]`.
    Otherwise, returns `None`.

    Parameters
    ----------
    ts
        None, a single `TimeSeries`, a sequence of `TimeSeries`, or a sequence of sequences of `TimeSeries`.

    Returns
    -------
    TimeSeries
        `ts` if  `ts` is a TimeSeries, `ts[0]` if `ts` is a Sequence of TimeSeries. Otherwise, returns `None`

    """
    seq_type = get_series_seq_type(ts)
    if seq_type <= SeriesType.SINGLE:
        # `None` and `TimeSeries`
        return ts
    elif seq_type == SeriesType.SEQ:
        return ts[0]
    else:
        return ts[0][0]


def get_series_seq_type(
    ts: Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]],
) -> SeriesType:
    """Returns the sequence type of `ts`.

    - SeriesType.SINGLE: `TimeSeries` (e.g. a single series)
    - SeriesType.SEQ: sequence of `TimeSeries` (e.g. multiple series)
    - SeriesType.SEQ_SEQ: sequence of sequences of `TimeSeries` (e.g. historical forecasts output)

    Parameters
    ----------
    ts
        The input series to get the sequence type from.

    Raises
    ------
    ValueError
        If `ts` does not have one of the expected sequence types.
    """
    if ts is None:
        return SeriesType.NONE
    elif isinstance(ts, TimeSeries):
        return SeriesType.SINGLE
    elif isinstance(ts[0], TimeSeries):
        return SeriesType.SEQ
    else:
        try:
            if isinstance(ts[0][0], TimeSeries):
                return SeriesType.SEQ_SEQ
            else:
                raise_log(
                    ValueError(
                        "input series must be of type `TimeSeries`, `Sequence[TimeSeries]`, or "
                        "`Sequence[Sequence[TimeSeries]]`."
                    ),
                    logger=logger,
                )
        except Exception as err:
            raise_log(
                ValueError(
                    "input series must be of type `TimeSeries`, `Sequence[TimeSeries]`, or "
                    f"`Sequence[Sequence[TimeSeries]]`. Raised: `{type(err).__name__}('{str(err)}')`"
                ),
                logger=logger,
            )


# TODO: we do not check the time index here
def retain_period_common_to_all(series: list[TimeSeries]) -> list[TimeSeries]:
    """
    Trims all series in the provided list, if necessary, so that the returned time series have
    a common span (corresponding to largest time sub-interval common to all series).

    Parameters
    ----------
    series
        The list of series to consider.

    Raises
    ------
    ValueError
        If no common time sub-interval exists

    Returns
    -------
    List[TimeSeries]
        A list of series, where each series have the same span
    """

    last_first = max(map(lambda s: s.start_time(), series))
    first_last = min(map(lambda s: s.end_time(), series))

    if last_first >= first_last:
        raise_log(
            ValueError("The provided time series must have nonzero overlap"), logger
        )

    return list(map(lambda s: s.slice(last_first, first_last), series))



import numpy as np
import pytest

from darts import TimeSeries
from darts.models import ExponentialSmoothing
from darts.utils import timeseries_generation as tg
from darts.utils.utils import freqs


class TestExponentialSmoothing:
    series = tg.sine_timeseries(length=100, freq=freqs["h"])

    @pytest.mark.parametrize(
        "freq_string,expected_seasonal_periods",
        [
            ("D", 7),
            (freqs["h"], 24),
            (freqs["ME"], 12),
            ("W", 52),
            (freqs["QE"], 4),
            ("B", 5),
        ],
    )
    def test_seasonality_inference(
        self, freq_string: str, expected_seasonal_periods: int
    ):
        series = tg.sine_timeseries(length=200, freq=freq_string)
        model = ExponentialSmoothing()
        model.fit(series)
        assert model.seasonal_periods == expected_seasonal_periods

    def test_default_parameters(self):
        """Test default selection for integer index"""
        series = TimeSeries.from_values(np.arange(1, 30, 1))
        model = ExponentialSmoothing()
        model.fit(series)
        assert model.seasonal_periods == 12

    def test_multiple_fit(self):
        """Test whether a model that inferred a seasonality period before will do it again for a new series"""
        series1 = tg.sine_timeseries(length=100, freq=freqs["ME"])
        series2 = tg.sine_timeseries(length=100, freq="D")
        model = ExponentialSmoothing()
        model.fit(series1)
        model.fit(series2)
        assert model.seasonal_periods == 7

    def test_constructor_kwargs(self):
        """Using kwargs to pass additional parameters to the constructor"""
        constructor_kwargs = {
            "initialization_method": "known",
            "initial_level": 0.5,
            "initial_trend": 0.2,
            "initial_seasonal": np.arange(1, 25),
        }
        model = ExponentialSmoothing(kwargs=constructor_kwargs)
        model.fit(self.series)
        # must be checked separately, name is not consistent
        np.testing.assert_array_almost_equal(
            model.model.model.params["initial_seasons"],
            constructor_kwargs["initial_seasonal"],
        )
        for param_name in ["initial_level", "initial_trend"]:
            assert (
                model.model.model.params[param_name] == constructor_kwargs[param_name]
            )

    def test_fit_kwargs(self):
        """Using kwargs to pass additional parameters to the fit()"""
        # using default optimization method
        model = ExponentialSmoothing()
        model.fit(self.series)
        assert model.fit_kwargs == {}
        pred = model.predict(n=2)

        model_bis = ExponentialSmoothing()
        model_bis.fit(self.series)
        assert model_bis.fit_kwargs == {}
        pred_bis = model_bis.predict(n=2)

        # two methods with the same parameters should yield the same forecasts
        assert pred.time_index.equals(pred_bis.time_index)
        np.testing.assert_array_almost_equal(pred.values(), pred_bis.values())

        # change optimization method
        model_ls = ExponentialSmoothing(method="least_squares")
        model_ls.fit(self.series)
        assert model_ls.fit_kwargs == {"method": "least_squares"}
        pred_ls = model_ls.predict(n=2)

        # forecasts should be slightly different
        assert pred.time_index.equals(pred_ls.time_index)
        assert all(np.not_equal(pred.values(), pred_ls.values()))



"""
Invertible Data Transformer Base Class
--------------------------------------
"""

from abc import abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, Optional, Union

import numpy as np

from darts import TimeSeries
from darts.dataprocessing.transformers.base_data_transformer import (
    BaseDataTransformer,
    component_masking,
)
from darts.logging import get_logger, raise_log
from darts.utils import _build_tqdm_iterator, _parallel_apply

logger = get_logger(__name__)


class InvertibleDataTransformer(BaseDataTransformer):
    def __init__(
        self,
        name: str = "InvertibleDataTransformer",
        n_jobs: int = 1,
        verbose: bool = False,
        parallel_params: Union[bool, Sequence[str]] = False,
        mask_components: bool = True,
    ):
        """Abstract class for invertible transformers.

        All the deriving classes have to implement the static methods :func:`ts_transform()` and
        :func:`ts_inverse_transform()`. For information on how to implement the :func:`ts_transform` method,
        please refer to the :class:`.BaseDataTransformer` documentation.

        The :func:`ts_inverse_transform()` method should be implemented in a virtually identical way to the
        :func:`ts_transform()` method: it should accept a ``TimeSeries`` as a first argument, and a dictionary
        of fixed parameters (as well as fitted parameters if the transformation also inherits from
        `FittableDataTransformer`) as a second argument. Additionally, :func:`ts_inverse_transform()` should
        also accept `*args` and `**kwargs` if additional positional/keyword arguments are expected to be
        passed. The only difference between :func:`ts_inverse_transform()` and :func:`ts_transform()`
        is that the former should 'undo' the transformation made to a `TimeSeries` by the latter. Please
        refer to the :func:`ts_inverse_transform()` documentation for further information.

        This class takes care of parallelizing the transformation on multiple ``TimeSeries`` when possible.

        Parameters
        ----------
        name
            The data transformer's name
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a Sequence[TimeSeries] is passed
            as input to a method, parallelising operations regarding different TimeSeries. Defaults to `1` (sequential).
            Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress
        parallel_params
            Optionally, specifies which fixed parameters (i.e. the attributes initialized in the child-most
            class's `__init__`) take on different values for different parallel jobs. Fixed parameters specified
            by `parallel_params` are assumed to be a `Sequence` of values that should be used for that parameter
            in each parallel job; the length of this `Sequence` should equal the number of parallel jobs. If
            `parallel_params=True`, every fixed parameter will take on a different value for each
            parallel job. If `parallel_params=False`, every fixed parameter will take on the same value for
            each parallel job. If `parallel_params` is a `Sequence` of fixed attribute names, only those
            attribute names specified will take on different values between different parallel jobs.
        mask_components
            Optionally, whether to automatically apply any provided `component_mask`s to the
            `TimeSeries` inputs passed to `transform`, `fit`, `inverse_transform`, or `fit_transform`.
            If `True`, any specified `component_mask` will be applied to each input timeseries
            before passing them to the called method; the masked components will also be automatically
            'unmasked' in the returned `TimeSeries`. If `False`, then `component_mask` (if provided) will
            be passed as a keyword argument, but won't automatically be applied to the input timeseries.
            See `apply_component_mask` method of `BaseDataTransformer` for further details.

        Notes
        -----
        Note: the :func:`ts_transform()` and :func:`ts_inverse_transform()` methods are designed to be
        static methods instead of instance methods to allow an efficient parallelisation also when the
        scaler instance is storing a non-negligible amount of data. Using instance methods would imply
        copying the instance's data through multiple processes, which can easily introduce a bottleneck
        and nullify parallelisation benefits.

        Example
        --------
        >>> from darts.dataprocessing.transformers import InvertibleDataTransformer
        >>> from darts.utils.timeseries_generation import linear_timeseries
        >>>
        >>> class SimpleTransform(InvertibleDataTransformer):
        >>>
        >>>         def __init__(self, a):
        >>>             self._a = a
        >>>             super().__init__()
        >>>
        >>>         @staticmethod
        >>>         def ts_transform(series, params, **kwargs):
        >>>             a = params['fixed']['_a']
        >>>             b = kwargs.pop('b')
        >>>             return a*series + b
        >>>
        >>>         @staticmethod
        >>>         def ts_inverse_transform(series, params, **kwargs):
        >>>             a = params['fixed']['_a']
        >>>             b = kwargs.pop('b')
        >>>             return (series - b) / a
        >>>
        >>> series = linear_timeseries(length=5)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[0.  ]],

            [[0.25]],

            [[0.5 ]],

            [[0.75]],

            [[1.  ]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        >>> transform = SimpleTransform(a=2)
        >>> series = transform.transform(series, b=3)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[3. ]],

            [[3.5]],

            [[4. ]],

            [[4.5]],

            [[5. ]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        >>> series = transform.inverse_transform(series, b=3)
        >>> print(series)
        <TimeSeries (DataArray) (time: 5, component: 1, sample: 1)>
        array([[[0.  ]],

            [[0.25]],

            [[0.5 ]],

            [[0.75]],

            [[1.  ]]])
        Coordinates:
        * time       (time) datetime64[ns] 2000-01-01 2000-01-02 ... 2000-01-05
        * component  (component) object 'linear'
        Dimensions without coordinates: sample
        Attributes:
            static_covariates:  None
            hierarchy:          None
        """
        super().__init__(
            name=name,
            n_jobs=n_jobs,
            verbose=verbose,
            parallel_params=parallel_params,
            mask_components=mask_components,
        )

    @classmethod
    @component_masking
    def _ts_inverse_transform(cls, *args, **kwargs):
        """Applies component masking to `ts_inverse_transform`."""
        return cls.ts_inverse_transform(*args, **kwargs)

    @staticmethod
    @abstractmethod
    def ts_inverse_transform(
        series: TimeSeries, params: Mapping[str, Any]
    ) -> TimeSeries:
        """The function that will be applied to each series when :func:`inverse_transform` is called.

        The function must take as first argument a ``TimeSeries`` object and, as a second argument, a
        dictionary containing the fixed and/or fitted parameters of the transformation; this function
        should then return an inverse transformed ``TimeSeries`` object (i.e. `ts_inverse_transform` should
        'undo' the transformation performed by `ts_transform`).

        The `params` dictionary *can* contain up to two keys:
            1. `params['fixed']` stores the fixed parameters of the transformation (i.e. attributed
            defined in the `__init__` method of the child-most class *before* `super().__init__` is called);
            `params['fixed']` is a dictionary itself, whose keys are the names of the fixed parameter
            attributes. For example, if `_my_fixed_param` is defined as an attribute in the child-most
            class, then this fixed parameter value can be accessed through `params['fixed']['_my_fixed_param']`.
            2. If the transform inherits from the :class:`.FittableDataTransformer` class, then `params['fitted']`
            will store the fitted parameters of the transformation; the fitted parameters are simply the output(s)
            returned by the `ts_fit` function, whatever those output(s) may be. See :class:`.FittableDataTransformer`
            for further details about fitted parameters.

        Any positional/keyword argument supplied to the `transform` method are passed as positional/keyword arguments
        to `ts_inverse_transform`; hence, `ts_inverse_transform` should also accept `*args` and/or `**kwargs` if
        positional/keyword arguments are passed to `transform`. Note that if the `mask_components` attribute of
        `InvertibleDataTransformer` is set to `False`, then the `component_mask` provided to `transform` will be passed
        as an additional keyword argument to `ts_inverse_transform`.

        The `BaseDataTransformer` class, from which `InvertibleDataTransformer` inherits, includes some helper methods
        which may prove useful when implementing a `ts_inverse_transform` function:
            1. The `apply_component_mask` and `unapply_component_mask` methods, which apply and 'unapply'
            `component_mask`s to a `TimeSeries` respectively; these methods are automatically called in `transform` if
            the `mask_component` attribute of `InvertibleDataTransformer` is set to `True`, but you may want to manually
            call them if you set `mask_components` to `False` and wish to manually specify how `component_mask`s are
            applied to a `TimeSeries`.
            2. The `stack_samples` method, which stacks all the samples in a `TimeSeries` along
            the component axis, so that the `TimeSeries` goes from shape `(n_timesteps, n_components, n_samples)` to
            shape `(n_timesteps, n_components * n_samples)`. This stacking is useful if a pointwise inverse transform
            is being implemented (i.e. transforming the value at time `t` depends only on the value of the series at
            that time `t`). Once transformed, the stacked `TimeSeries` can be 'unstacked' using the `unstack_samples`
            method.

        This method is not implemented in the base class and must be implemented in the deriving classes.

        Parameters
        ----------
        series
            series to be transformed.
        params
            Dictionary containing the parameters of the transformation function. Fixed parameters
            (i.e. attributes defined in the child-most class of the transformation prior to
            calling `super.__init__()`) are stored under the `'fixed'` key. If the transformation
            inherits from the `FittableDataTransformer` class, then the fitted parameters of the
            transformation (i.e. the values returned by `ts_fit`) are stored under the
            `'fitted'` key.
        args
            Any additional keyword arguments provided to `inverse_transform`.
        kwargs
            Any additional keyword arguments provided to `inverse_transform`. Note that if the `mask_component`
            attribute of `InvertibleDataTransformer` is set to `False`, then `component_mask` will
            be passed as a keyword argument.

        Notes
        -----
        This method is designed to be a static method instead of instance methods to allow an efficient
        parallelisation also when the scaler instance is storing a non-negligible amount of data. Using instance
        methods would imply copying the instance's data through multiple processes, which can easily introduce a
        bottleneck and nullify parallelisation benefits.
        """
        pass

    def inverse_transform(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]],
        *args,
        component_mask: Optional[np.array] = None,
        series_idx: Optional[Union[int, Sequence[int]]] = None,
        **kwargs,
    ) -> Union[TimeSeries, list[TimeSeries], list[list[TimeSeries]]]:
        """Inverse transforms a (sequence of) series by calling the user-implemented `ts_inverse_transform` method.

        In case a sequence or list of lists is passed as input data, this function takes care of parallelising the
        transformation of multiple series in the sequence at the same time. Additionally,
        if the `mask_components` attribute was set to `True` when instantiating `InvertibleDataTransformer`,
        then any provided `component_mask`s will be automatically applied to each input `TimeSeries`;
        please refer to 'Notes' for further details on component masking.

        Any additionally specified `*args` and `**kwargs` are automatically passed to `ts_inverse_transform`.

        Parameters
        ----------
        series
            The series to inverse-transform.
            If a single `TimeSeries`, returns a single series.
            If a sequence of `TimeSeries`, returns a list of series. The series should be in the same order as the
            sequence used to fit the transformer.
            If a list of lists of `TimeSeries`, returns a list of lists of series. This can for example be the output
            of `ForecastingModel.historical_forecasts()` when using multiple series. Each inner list should contain
            `TimeSeries` related to the same series. The order of inner lists should be the same as the sequence used
            to fit the transformer.
        args
            Additional positional arguments for the :func:`ts_inverse_transform()` method
        component_mask : Optional[np.ndarray] = None
            Optionally, a 1-D boolean np.ndarray of length ``series.n_components`` that specifies
            which components of the underlying `series` the inverse transform should consider.
        series_idx
            Optionally, the index(es) of each series corresponding to their positions within the series used to fit
            the transformer (to retrieve the appropriate transformer parameters).
        kwargs
            Additional keyword arguments for the :func:`ts_inverse_transform()` method

        Returns
        -------
        Union[TimeSeries, List[TimeSeries], List[List[TimeSeries]]]
            Inverse transformed data.

        Notes
        -----
        If the `mask_components` attribute was set to `True` when instantiating `InvertibleDataTransformer`,
        then any provided `component_mask`s will be automatically applied to each `TimeSeries` input to
        transform; `component_mask`s are simply boolean arrays of shape `(series.n_components,)` that
        specify which components of each `series` should be transformed using `ts_inverse_transform` and which
        components should not. If `component_mask[i]` is `True`, then the `i`th component of each
        `series` will be transformed by `ts_inverse_transform`. Conversely, if `component_mask[i]` is `False`,
        the `i`th component will be removed from each `series` before being passed to `ts_inverse_transform`;
        after transforming this masked series, the untransformed `i`th component will be 'added back'
        to the output. Note that automatic `component_mask`ing can only be performed if the `ts_inverse_transform`
        does *not* change the number of timesteps in each series; if this were to happen, then the transformed
        and untransformed components are unable to be concatenated back together along the component axis.

        If `mask_components` was set to `False` when instantiating `InvertibleDataTransformer`, then any provided
        `component_masks` will be passed as a keyword argument `ts_inverse_transform`; the user can then manually
        specify how the `component_mask` should be applied to each series.
        """
        if hasattr(self, "_fit_called") and not self._fit_called:
            raise_log(
                ValueError("fit() must have been called before inverse_transform()"),
                logger=logger,
            )

        desc = f"Inverse ({self._name})"

        # Take note of original input for unmasking purposes:
        called_with_single_series = False
        called_with_sequence_series = False
        series_specified = series_idx is not None
        if isinstance(series, TimeSeries):
            data = [series]
            if series_specified:
                transformer_selector = self._process_series_idx(series_idx)
            else:
                transformer_selector = [0]
            called_with_single_series = True
        elif isinstance(series[0], TimeSeries):  # Sequence[TimeSeries]
            data = series
            if series_specified:
                transformer_selector = self._process_series_idx(series_idx)
            else:
                transformer_selector = range(len(series))
            called_with_sequence_series = True
        else:  # Sequence[Sequence[TimeSeries]]
            data = []
            transformer_selector = []
            if series_specified:
                iterator_ = zip(self._process_series_idx(series_idx), series)
            else:
                iterator_ = enumerate(series)
            for idx, series_list in iterator_:
                data.extend(series_list)
                transformer_selector += [idx] * len(series_list)

        input_iterator = _build_tqdm_iterator(
            zip(
                data,
                self._get_params(
                    transformer_selector=transformer_selector,
                    series_specified=series_specified,
                ),
            ),
            verbose=self._verbose,
            desc=desc,
            total=len(transformer_selector),
        )

        # apply & unapply component masking to the transform method
        kwargs["mask_components"] = self._mask_components
        kwargs["mask_components_apply_only"] = False
        kwargs["component_mask"] = component_mask

        transformed_data = _parallel_apply(
            input_iterator,
            self._ts_inverse_transform,
            self._n_jobs,
            args,
            kwargs,
        )

        if called_with_single_series:
            return transformed_data[0]
        elif called_with_sequence_series:
            return transformed_data
        else:
            cum_len = np.cumsum([0] + [len(s_) for s_ in series])
            return [
                transformed_data[cum_len[i] : cum_len[i + 1]]
                for i in range(len(cum_len) - 1)
            ]



"""
Metrics
-------

For deterministic forecasts (point predictions with `num_samples == 1`), probabilistic forecasts (`num_samples > 1`),
and quantile forecasts. For probabilistic and quantile forecasts, use parameter `q` to define the quantile(s) to
compute the deterministic metrics on:

- Aggregated over time:
    Absolute metrics:
        - :func:`MERR <darts.metrics.metrics.merr>`: Mean Error
        - :func:`MAE <darts.metrics.metrics.mae>`: Mean Absolute Error
        - :func:`MSE <darts.metrics.metrics.mse>`: Mean Squared Error
        - :func:`RMSE <darts.metrics.metrics.rmse>`: Root Mean Squared Error
        - :func:`RMSLE <darts.metrics.metrics.rmsle>`: Root Mean Squared Log Error

    Relative metrics:
        - :func:`MASE <darts.metrics.metrics.mase>`: Mean Absolute Scaled Error
        - :func:`MSSE <darts.metrics.metrics.msse>`: Mean Squared Scaled Error
        - :func:`RMSSE <darts.metrics.metrics.rmsse>`: Root Mean Squared Scaled Error
        - :func:`MAPE <darts.metrics.metrics.mape>`: Mean Absolute Percentage Error
        - :func:`wMAPE <darts.metrics.metrics.wmape>`: weighted Mean Absolute Percentage Error
        - :func:`sMAPE <darts.metrics.metrics.smape>`: symmetric Mean Absolute Percentage Error
        - :func:`OPE <darts.metrics.metrics.ope>`: Overall Percentage Error
        - :func:`MARRE <darts.metrics.metrics.marre>`: Mean Absolute Ranged Relative Error

    Other metrics:
        - :func:`R2 <darts.metrics.metrics.r2_score>`: Coefficient of Determination
        - :func:`CV <darts.metrics.metrics.coefficient_of_variation>`: Coefficient of Variation

- Per time step:
    Absolute metrics:
        - :func:`ERR <darts.metrics.metrics.err>`: Error
        - :func:`AE <darts.metrics.metrics.ae>`: Absolute Error
        - :func:`SE <darts.metrics.metrics.se>`: Squared Error
        - :func:`SLE <darts.metrics.metrics.sle>`: Squared Log Error

    Relative metrics:
        - :func:`ASE <darts.metrics.metrics.ase>`: Absolute Scaled Error
        - :func:`SSE <darts.metrics.metrics.sse>`: Squared Scaled Error
        - :func:`APE <darts.metrics.metrics.ape>`: Absolute Percentage Error
        - :func:`sAPE <darts.metrics.metrics.sape>`: symmetric Absolute Percentage Error
        - :func:`ARRE <darts.metrics.metrics.arre>`: Absolute Ranged Relative Error

For probabilistic forecasts (storchastic predictions with `num_samples >> 1`) and quantile forecasts:

- Aggregated over time:
    Quantile metrics:
        - :func:`MQL <darts.metrics.metrics.mql>`: Mean Quantile Loss
        - :func:`QR <darts.metrics.metrics.qr>`: Quantile Risk

    Quantile interval metrics:
        - :func:`MIW <darts.metrics.metrics.miw>`: Mean Interval Width
        - :func:`MWS <darts.metrics.metrics.miws>`: Mean Interval Winkler Score
        - :func:`MIC <darts.metrics.metrics.mic>`: Mean Interval Coverage
        - :func:`MINCS_QR <darts.metrics.metrics.mincs_qr>`: Mean Interval Non-Conformity Score for Quantile Regression

- Per time step:
    Quantile metrics:
        - :func:`QL <darts.metrics.metrics.ql>`: Quantile Loss

    Quantile interval metrics:
        - :func:`IW <darts.metrics.metrics.iw>`: Interval Width
        - :func:`WS <darts.metrics.metrics.iws>`: Interval Winkler Score
        - :func:`IC <darts.metrics.metrics.ic>`: Interval Coverage
        - :func:`INCS_QR <darts.metrics.metrics.incs_qr>`: Interval Non-Conformity Score for Quantile Regression

For Dynamic Time Warping (DTW) (aggregated over time):

- :func:`DTW <darts.metrics.metrics.dtw_metric>`: Dynamic Time Warping Metric
"""

from darts.metrics.metrics import (
    ae,
    ape,
    arre,
    ase,
    coefficient_of_variation,
    dtw_metric,
    err,
    ic,
    incs_qr,
    iw,
    iws,
    mae,
    mape,
    marre,
    mase,
    merr,
    mic,
    mincs_qr,
    miw,
    miws,
    mql,
    mse,
    msse,
    ope,
    ql,
    qr,
    r2_score,
    rmse,
    rmsle,
    rmsse,
    sape,
    se,
    sle,
    smape,
    sse,
    wmape,
)

ALL_METRICS = {
    ae,
    ape,
    arre,
    ase,
    coefficient_of_variation,
    dtw_metric,
    err,
    iw,
    iws,
    mae,
    mape,
    wmape,
    marre,
    mase,
    merr,
    miw,
    miws,
    mql,
    mse,
    msse,
    ope,
    ql,
    qr,
    r2_score,
    rmse,
    rmsle,
    rmsse,
    sape,
    se,
    sle,
    smape,
    sse,
    ic,
    mic,
    incs_qr,
    mincs_qr,
}

TIME_DEPENDENT_METRICS = {
    ae,
    ape,
    arre,
    ase,
    err,
    ql,
    sape,
    se,
    sle,
    sse,
    iw,
    iws,
    ic,
    incs_qr,
}

Q_INTERVAL_METRICS = {
    iw,
    iws,
    miw,
    miws,
    ic,
    mic,
    incs_qr,
}

NON_Q_METRICS = {dtw_metric}

__all__ = [
    "ae",
    "ape",
    "arre",
    "ase",
    "coefficient_of_variation",
    "dtw_metric",
    "err",
    "mae",
    "mape",
    "wmape",
    "marre",
    "mase",
    "merr",
    "mql",
    "mse",
    "msse",
    "ope",
    "ql",
    "qr",
    "r2_score",
    "rmse",
    "rmsle",
    "rmsse",
    "sape",
    "se",
    "sle",
    "smape",
    "sse",
    "iw",
    "miw",
    "iws",
    "miws",
    "ic",
    "mic",
    "incs_qr",
    "mincs_qr",
]



"""
Utils for filling missing values
--------------------------------
"""

from typing import Optional, Union

from darts.logging import get_logger, raise_if, raise_if_not
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


def missing_values_ratio(series: TimeSeries) -> float:
    """
    Computes the ratio of missing values

    Parameters
    ----------
    series
        The time series to compute ratio on

    Returns
    -------
    float
        The ratio of missing values
    """

    return series.to_dataframe().isnull().sum().mean() / len(series)


def fill_missing_values(
    series: TimeSeries, fill: Union[str, float] = "auto", **interpolate_kwargs
) -> TimeSeries:
    """
    Fills missing values in the provided time series

    Parameters
    ----------
    series
        The time series for which to fill missing values
    fill
        The value used to replace the missing values.
        If set to 'auto', will auto-fill missing values using the `pandas.Dataframe.interpolate()` method.
    interpolate_kwargs
        Keyword arguments for `pandas.Dataframe.interpolate()`, only used when fit is set to 'auto'.
        See `the documentation
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.interpolate.html>`_
        for the list of supported parameters.

    Returns
    -------
    TimeSeries
        A new TimeSeries with all missing values filled according to the rules above.
    """
    raise_if_not(
        isinstance(fill, str) or isinstance(fill, float),
        "`fill` should either be a string or a float",
        logger,
    )
    raise_if(
        isinstance(fill, str) and fill != "auto",
        "invalid string for `fill`: can only be set to 'auto'",
        logger,
    )

    if fill == "auto":
        return _auto_fill(series, **interpolate_kwargs)
    return _const_fill(series, fill)


def extract_subseries(
    series: TimeSeries, min_gap_size: Optional[int] = 1, mode: str = "all"
) -> list[TimeSeries]:
    """
    Partitions the series into a sequence of sub-series by using significant gaps of missing values

    Parameters
    ----------
    series
        The TimeSeries to partition into sub-series

    min_gap_size
        The minimum number of contiguous missing values to consider a gap as significant. Defaults to 1.

    mode
        Only for multivariate TimeSeries. The definition of a gap; presence of a NaN in any column ("any")
        or NaNs in all the columns ("all") for a given timestamp. Defaults to "all".

    Returns
    -------
    subseries
        A list of TimeSeries, sub-series without significant gaps of missing values

    See Also
    --------
    TimeSeries.gaps : return the gaps in the TimeSeries
    """

    # Remove null values from the series extremes
    series = series.strip()
    freq = series.freq

    if series.to_dataframe().isna().sum().sum() == 0:
        return [series]

    # Get start/end times of sub-series without gaps of missing values
    gaps_df = series.gaps(mode=mode)
    if gaps_df.empty:
        return series
    else:
        gaps_df = gaps_df.query(f"gap_size>={min_gap_size}")
        start_times = [series.start_time()] + (gaps_df["gap_end"] + freq).to_list()
        end_times = (gaps_df["gap_start"] - freq).to_list() + [series.end_time() + freq]

        subseries = []
        for start, end in zip(start_times, end_times):
            subseries.append(series[start:end])

        return subseries


def _const_fill(series: TimeSeries, fill: float = 0) -> TimeSeries:
    """
    Fills the missing values of `series` with only the value provided (default zeroes).

    Parameters
    ----------
    series
        The TimeSeries to check for missing values.
    fill
        The value used to replace the missing values.

    Returns
    -------
    TimeSeries
        A TimeSeries, `series` with all missing values set to `fill`.
    """

    return TimeSeries.from_times_and_values(
        series.time_index,
        series.to_dataframe().fillna(value=fill),
        freq=series.freq,
        columns=series.columns,
        static_covariates=series.static_covariates,
        hierarchy=series.hierarchy,
        metadata=series.metadata,
    )


def _auto_fill(series: TimeSeries, **interpolate_kwargs) -> TimeSeries:
    """
    This function fills the missing values in the TimeSeries `series`,
    using the `pandas.Dataframe.interpolate()` method.

    Parameters
    ----------
    series
        The time series
    interpolate_kwargs
        Keyword arguments for `pandas.Dataframe.interpolate()`.
        See `the documentation
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.interpolate.html>`_
        for the list of supported parameters.
    Returns
    -------
    TimeSeries
        A new TimeSeries with all missing values filled according to the rules above.
    """

    series_temp = series.to_dataframe()

    # pandas interpolate wrapper, with chosen `method`
    if "limit_direction" not in interpolate_kwargs:
        interpolate_kwargs["limit_direction"] = "both"
    interpolate_kwargs["inplace"] = True
    series_temp.interpolate(**interpolate_kwargs)
    return TimeSeries.from_dataframe(
        series_temp,
        freq=series.freq,
        static_covariates=series.static_covariates,
        hierarchy=series.hierarchy,
        metadata=series.metadata,
    )
